/*
 * named_pipe.c
 *
 * Copyright (C) 2006  Insigme Co., Ltd
 *
 * This software has been developed while working on the Linux Unified Kernel
 * project (http://www.longene.org) in the Insigma Research Institute,  
 * which is a subdivision of Insigma Co., Ltd (http://www.insigma.com.cn).
 * 
 * The project is sponsored by Insigma Co., Ltd.
 *
 * The authors can be reached at linux@insigma.com.cn.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of  the GNU General  Public License as published by the
 * Free Software Foundation; either version 2 of the  License, or (at your
 * option) any later version.
 *
 * Revision History:
 *   Dec 2008 - Created.
 */
 
/*
 * named_pipe.c:
 * Refered to Wine code
 */
#include <linux/mm.h>
#include <linux/syscalls.h>

#include "wineserver/server.h"
#include "wineserver/request.h"
#include "wineserver/uk_lib.h"
#include "io.h"
#include "object.h"
#include "unistr.h"
#include "event.h"
#include "object.h"
#include "section.h"
#include "handle.h"
#include "objwait.h"
#include "winuser.h"
#include "wineserver/winerror.h"
#include "thread.h"
#include "wineserver/file.h"
#include "wineserver/user.h"
#include "wineserver/winerror.h"
#include "file.h"

#ifdef CONFIG_UNIFIED_KERNEL
enum pipe_state
{
    ps_idle_server,
    ps_wait_open,
    ps_connected_server,
    ps_wait_disconnect,
    ps_disconnected_server,
    ps_wait_connect
};

struct named_pipe;

struct pipe_server
{
    struct object        obj;        /* object header */
    struct fd           *fd;         /* pipe file descriptor */
    struct fd           *ioctl_fd;   /* file descriptor for ioctls when not connected */
    struct list_head          entry;      /* entry in named pipe servers list */
    enum pipe_state      state;      /* server state */
    struct pipe_client  *client;     /* client that this server is connected to */
    struct named_pipe   *pipe;
    struct timeout_user *flush_poll;
    struct kevent        *event;
    unsigned int         options;    /* pipe options */
};

struct pipe_client
{
    struct object        obj;        /* object header */
    struct fd           *fd;         /* pipe file descriptor */
    struct pipe_server  *server;     /* server that this client is connected to */
    unsigned int         flags;      /* file flags */
};

struct named_pipe
{
    struct object       obj;         /* object header */
    unsigned int        flags;
    unsigned int        maxinstances;
    unsigned int        outsize;
    unsigned int        insize;
    unsigned int        instances;
    timeout_t           timeout;
    struct list_head         servers;     /* list of servers using this pipe */
    struct async_queue *waiters;     /* list of clients waiting to connect */
};

struct named_pipe_device
{
    struct object       obj;         /* object header */
    struct fd           *fd;         /* pseudo-fd for ioctls */
#if 0
    struct namespace   *pipes;       /* named pipe namespace */
#endif
    obj_handle_t        *pipes;      /* named pipe namespace */
};

static void named_pipe_dump( struct object *obj, int verbose );
static unsigned int named_pipe_map_access( struct object *obj, unsigned int access );
static struct object *named_pipe_open_file( struct object *obj, unsigned int access,
                                            unsigned int sharing, unsigned int options );
static void named_pipe_destroy( struct object *obj );

static const struct object_ops named_pipe_ops =
{
    sizeof(struct named_pipe),    /* size */
    named_pipe_dump,              /* dump */
    no_get_type,                  /* get_type */
    no_get_fd,                    /* get_fd */
    named_pipe_map_access,        /* map_access */
    no_lookup_name,               /* lookup_name */
    named_pipe_open_file,         /* open_file */
    no_close_handle,              /* close_handle */
    named_pipe_destroy            /* destroy */
};

/* server end functions */
static void pipe_server_dump( struct object *obj, int verbose );
static struct fd *pipe_server_get_fd( struct object *obj );
static void pipe_server_destroy( struct object *obj);
static void pipe_server_flush( struct fd *fd, struct kevent **event );
static enum server_fd_type pipe_server_get_fd_type( struct fd *fd );
static obj_handle_t pipe_server_ioctl( struct fd *fd, ioctl_code_t code, const async_data_t *async,
                                       const void *data, data_size_t size );
extern unsigned int default_fd_map_access( struct object *obj, unsigned int access );

static const struct object_ops pipe_server_ops =
{
    sizeof(struct pipe_server),   /* size */
    pipe_server_dump,             /* dump */
    no_get_type,                  /* get_type */
    pipe_server_get_fd,           /* get_fd */
    default_fd_map_access,        /* map_access */
    no_lookup_name,               /* lookup_name */
    no_open_file,                 /* open_file */
    fd_close_handle,              /* close_handle */
    pipe_server_destroy           /* destroy */
};

static const struct fd_ops pipe_server_fd_ops =
{
    default_fd_get_poll_events,   /* get_poll_events */
    default_poll_event,           /* poll_event */
    pipe_server_flush,            /* flush */
    pipe_server_get_fd_type,      /* get_fd_type */
    pipe_server_ioctl,            /* ioctl */
    default_fd_queue_async,       /* queue_async */
    default_fd_reselect_async,    /* reselect_async */
    default_fd_cancel_async,      /* cancel_async */
};

/* client end functions */
static void pipe_client_dump( struct object *obj, int verbose );
static struct fd *pipe_client_get_fd( struct object *obj );
static void pipe_client_destroy( struct object *obj );
static void pipe_client_flush( struct fd *fd, struct kevent **event );
static enum server_fd_type pipe_client_get_fd_type( struct fd *fd );

static const struct object_ops pipe_client_ops =
{
    sizeof(struct pipe_client),   /* size */
    pipe_client_dump,             /* dump */
    no_get_type,                  /* get_type */
    pipe_client_get_fd,           /* get_fd */
    default_fd_map_access,        /* map_access */
    no_lookup_name,               /* lookup_name */
    no_open_file,                 /* open_file */
    fd_close_handle,              /* close_handle */
    pipe_client_destroy           /* destroy */
};

static const struct fd_ops pipe_client_fd_ops =
{
    default_fd_get_poll_events,   /* get_poll_events */
    default_poll_event,           /* poll_event */
    pipe_client_flush,            /* flush */
    pipe_client_get_fd_type,      /* get_fd_type */
    default_fd_ioctl,             /* ioctl */
    default_fd_queue_async,       /* queue_async */
    default_fd_reselect_async,    /* reselect_async */
    default_fd_cancel_async       /* cancel_async */
};

static void named_pipe_device_dump( struct object *obj, int verbose );
static struct object_type *named_pipe_device_get_type( struct object *obj );
static struct fd *named_pipe_device_get_fd( struct object *obj );
static struct object *named_pipe_device_lookup_name( struct object *obj,
    struct unicode_str *name, unsigned int attr );
static struct object *named_pipe_device_open_file( struct object *obj, unsigned int access,
                                                   unsigned int sharing, unsigned int options );
static void named_pipe_device_destroy( struct object *obj );
static enum server_fd_type named_pipe_device_get_fd_type( struct fd *fd );
static obj_handle_t named_pipe_device_ioctl( struct fd *fd, ioctl_code_t code, const async_data_t *async_data,
                                             const void *data, data_size_t size );

static const struct object_ops named_pipe_device_ops =
{
    sizeof(struct named_pipe_device), /* size */
    named_pipe_device_dump,           /* dump */
    named_pipe_device_get_type,       /* get_type */
    named_pipe_device_get_fd,         /* get_fd */
    no_map_access,                    /* map_access */
    named_pipe_device_lookup_name,    /* lookup_name */
    named_pipe_device_open_file,      /* open_file */
    fd_close_handle,                  /* close_handle */
    named_pipe_device_destroy         /* destroy */
};

static const struct fd_ops named_pipe_device_fd_ops =
{
    default_fd_get_poll_events,       /* get_poll_events */
    default_poll_event,               /* poll_event */
    no_flush,                         /* flush */
    named_pipe_device_get_fd_type,    /* get_fd_type */
    named_pipe_device_ioctl,          /* ioctl */
    default_fd_queue_async,           /* queue_async */
    default_fd_reselect_async,        /* reselect_async */
    default_fd_cancel_async           /* cancel_async */
};

static NTSTATUS
parse_named_pipe(
		IN PVOID ParseObject,
		IN PVOID ObjectType,
		IN OUT PACCESS_STATE AccessState,
		IN KPROCESSOR_MODE AccessMode,
		IN ULONG Attributes,
		IN OUT PUNICODE_STRING CompleteName,
		IN OUT PUNICODE_STRING RemainingName,
		IN OUT PVOID Context OPTIONAL,
		IN PSECURITY_QUALITY_OF_SERVICE SecurityQos OPTIONAL,
		OUT PVOID *Object
		);

#define	NAMED_PIPE_ALL_ACCESS	(STANDARD_RIGHTS_REQUIRED | 0x1)
static WCHAR NamedPipeName[] = {'N', 'a', 'm', 'e', 'd', 'P', 'i', 'p', 'e', 0};
static WCHAR dir_named_pipe[] = {'\\', 'D', 'e', 'v', 'i', 'c', 'e', '\\', 'N', 'a', 'm', 'e', 'd', 'P', 'i', 'p', 'e', 0};
static WCHAR link_pipe[] = {'\\', '?', '?', '\\', 'P', 'I', 'P', 'E', 0};
static WCHAR named_pipe[] = {'N', 'a', 'm', 'e', 'd', 'P', 'i', 'p', 'e', 0};
extern HANDLE device_handle;
POBJECT_TYPE namedpipe_object_type;

void init_named_pipe(void)
{
	OBJECT_TYPE_INITIALIZER ObjectTypeInitializer;
	UNICODE_STRING Name, LinkName;
	OBJECT_ATTRIBUTES obj_attr;
	PVOID object;
	NTSTATUS status;

	ktrace("init_named_pipe\n");
	/* Initialize the File object type  */
	memset(&ObjectTypeInitializer, 0, sizeof(ObjectTypeInitializer));
	init_unistr(&Name, (PWSTR)NamedPipeName);
	ObjectTypeInitializer.Length = sizeof(ObjectTypeInitializer);
	ObjectTypeInitializer.DefaultNonPagedPoolCharge = sizeof(struct named_pipe_device);
	ObjectTypeInitializer.ParseProcedure = parse_named_pipe;
	create_type_object(&ObjectTypeInitializer, &Name, &namedpipe_object_type);

	init_unistr(&Name, (PWSTR)named_pipe);
	INIT_OBJECT_ATTR(&obj_attr, &Name, 0, device_handle, NULL);

	status = create_object(KernelMode,
					namedpipe_object_type,
					&obj_attr,
					KernelMode,
					NULL,
					named_pipe_device_ops.size,
					0,
					0,
					(PVOID *)&object);
	if (!NT_SUCCESS(status))
		return;

	status = insert_object(object,
					NULL,
					0,
					0,
					NULL,
					NULL);
	if (!NT_SUCCESS(status))
		return;

	BODY_TO_HEADER(object)->ops = &named_pipe_device_ops;

	init_unistr(&Name, (PWSTR)dir_named_pipe);
	init_unistr(&LinkName, (PWSTR)link_pipe);
	io_create_symbol_link(&LinkName, &Name);
}

static void named_pipe_dump( struct object *obj, int verbose )
{
#if 0
    struct named_pipe *pipe = (struct named_pipe *) obj;
    assert( obj->ops == &named_pipe_ops );
    fprintf( stderr, "Named pipe " );
    dump_object_name( &pipe->obj );
    fprintf( stderr, "\n" );
#endif
}

static unsigned int named_pipe_map_access( struct object *obj, unsigned int access )
{
    if (access & GENERIC_READ)
        access |= STANDARD_RIGHTS_READ;
    if (access & GENERIC_WRITE)
        access |= STANDARD_RIGHTS_WRITE | FILE_CREATE_PIPE_INSTANCE;
    if (access & GENERIC_EXECUTE)
        access |= STANDARD_RIGHTS_EXECUTE;
    if (access & GENERIC_ALL)
        access |= STANDARD_RIGHTS_ALL;
    return access & ~(GENERIC_READ | GENERIC_WRITE | GENERIC_EXECUTE | GENERIC_ALL);
}

static void pipe_server_dump( struct object *obj, int verbose )
{
#if 0
    struct pipe_server *server = (struct pipe_server *) obj;
    assert( obj->ops == &pipe_server_ops );
    fprintf( stderr, "Named pipe server pipe=%p state=%d\n", server->pipe, server->state );
#endif
}

static void pipe_client_dump( struct object *obj, int verbose )
{
#if 0
    struct pipe_client *client = (struct pipe_client *) obj;
    assert( obj->ops == &pipe_client_ops );
    fprintf( stderr, "Named pipe client server=%p\n", client->server );
#endif
}

static void named_pipe_destroy( struct object *obj)
{
    struct named_pipe *pipe = (struct named_pipe *) obj;

#if 0
    assert( list_empty( &pipe->servers ) );
    assert( !pipe->instances );
#endif
    free_async_queue( pipe->waiters );
}

static struct fd *pipe_client_get_fd( struct object *obj )
{
    struct pipe_client *client = (struct pipe_client *) obj;
    if (client->fd)
        return (struct fd *) grab_object( client->fd );
    set_error( STATUS_PIPE_DISCONNECTED );
    return NULL;
}

static void set_server_state( struct pipe_server *server, enum pipe_state state )
{
    server->state = state;

    switch(state) {
    case ps_connected_server:
    case ps_wait_disconnect:
#if 0
        assert( server->fd );
#endif
        break;
    case ps_wait_open:
    case ps_idle_server:
#if 0
        assert( !server->fd );
#endif
        set_no_fd_status( server->ioctl_fd, STATUS_PIPE_LISTENING );
        break;
    case ps_disconnected_server:
    case ps_wait_connect:
#if 0
        assert( !server->fd );
#endif
        set_no_fd_status( server->ioctl_fd, STATUS_PIPE_DISCONNECTED );
        break;
    }
}

static struct fd *pipe_server_get_fd( struct object *obj )
{
    struct pipe_server *server = (struct pipe_server *) obj;

    return (struct fd *)grab_object( server->fd ? server->fd : server->ioctl_fd );
}


static void notify_empty( struct pipe_server *server )
{
    if (!server->flush_poll)
        return;
#if 0
    assert( server->state == ps_connected_server );
    assert( server->event );
#endif
    remove_timeout_user( server->flush_poll );
    server->flush_poll = NULL;
    set_event( server->event, EVENT_INCREMENT, FALSE );
    release_object( server->event );
    server->event = NULL;
}

extern int uk_shutdown(int s, int how);

static void do_disconnect( struct pipe_server *server )
{
    /* we may only have a server fd, if the client disconnected */
    if (server->client) {
#if 0
        assert( server->client->server == server );
        assert( server->client->fd );
#endif
        release_object( server->client->fd );
        server->client->fd = NULL;
    }
#if 0
    assert( server->fd );
#endif
    uk_shutdown( get_unix_fd( server->fd ), SHUT_RDWR );
    release_object( server->fd );
    server->fd = NULL;
}

static void pipe_server_destroy( struct object *obj)
{
    struct pipe_server *server = (struct pipe_server *)obj;

#if 0
    assert( obj->ops == &pipe_server_ops );
#endif

    if (server->fd) {
        notify_empty( server );
        do_disconnect( server );
    }

    if (server->client) {
        server->client->server = NULL;
        server->client = NULL;
    }

#if 0
    assert( server->pipe->instances );
#endif
    server->pipe->instances--;

    if (server->ioctl_fd)
        release_object( server->ioctl_fd );
    list_remove( &server->entry );
    release_object( server->pipe );
}

static void pipe_client_destroy( struct object *obj)
{
    struct pipe_client *client = (struct pipe_client *)obj;
    struct pipe_server *server = client->server;

#if 0
    assert( obj->ops == &pipe_client_ops );
#endif

    if (server) {
        notify_empty( server );

        switch(server->state) {
        case ps_connected_server:
            /* Don't destroy the server's fd here as we can't
               do a successful flush without it. */
            set_server_state( server, ps_wait_disconnect );
            break;
        case ps_disconnected_server:
            set_server_state( server, ps_wait_connect );
            break;
        case ps_idle_server:
        case ps_wait_open:
        case ps_wait_disconnect:
        case ps_wait_connect:
#if 0
            assert( 0 );       /* D.M. TBD */
#endif
            ;
        }
#if 0
        assert( server->client );
#endif
        server->client = NULL;
        client->server = NULL;
    }
    if (client->fd)
        release_object( client->fd );
}

static void named_pipe_device_dump( struct object *obj, int verbose )
{
#if 0
    assert( obj->ops == &named_pipe_device_ops );
    fprintf( stderr, "Named pipe device\n" );
#endif
}

extern struct object_type *get_object_type( const struct unicode_str* );

static struct object_type *named_pipe_device_get_type( struct object *obj )
{
    static const WCHAR name[] = {'D','e','v','i','c','e'};
    static const struct unicode_str str = { name, sizeof(name) };
    return get_object_type( &str );
}

static struct fd *named_pipe_device_get_fd( struct object *obj )
{
    struct named_pipe_device *device = (struct named_pipe_device *)obj;
    return (struct fd *)grab_object( device->fd );
}

static struct object *named_pipe_device_lookup_name( struct object *obj, struct unicode_str *name,
                                                     unsigned int attr )
{
#if 0
    struct named_pipe_device *device = (struct named_pipe_device*)obj;
    struct object *found;

    assert( obj->ops == &named_pipe_device_ops );
    assert( device->pipes );

    if ((found = find_object( device->pipes, name, attr | OBJ_CASE_INSENSITIVE )))
        name->len = 0;

    return found;
#endif
    return NULL;
}

static struct object *named_pipe_device_open_file( struct object *obj, unsigned int access,
                                                   unsigned int sharing, unsigned int options )
{
    return grab_object( obj );
}

static void named_pipe_device_destroy( struct object *obj )
{
    struct named_pipe_device *device = (struct named_pipe_device*)obj;

#if 0
    assert( obj->ops == &named_pipe_device_ops );
#endif
    if (device->fd)
        release_object( device->fd );
    NtClose(device->pipes);
#if 0
    free( device->pipes );
#endif
}

static enum server_fd_type named_pipe_device_get_fd_type( struct fd *fd )
{
    return FD_TYPE_DEVICE;
}

static WCHAR PipeRoot[] = {'P', 'i', 'p', 'e', 'R', 'o', 'o', 't', 0};
static POBJECT_DIRECTORY PipeNameSpace;

void create_named_pipe_device( struct directory *root, const struct unicode_str *name )
{
    struct named_pipe_device *dev;
    UNICODE_STRING pipe_root;
    OBJECT_ATTRIBUTES ObjectAttributes;

    if ((dev = create_named_object_dir( root, name, 0, &named_pipe_device_ops )) &&
        get_error() != STATUS_OBJECT_NAME_EXISTS) {
       init_unistr(&pipe_root, (PWSTR)PipeRoot);
       INIT_OBJECT_ATTR(&ObjectAttributes,
                       &pipe_root,
                       0,
                       NULL,
                       NULL);
       create_object(KernelMode,
                       dir_object_type,
                       &ObjectAttributes,
                       KernelMode,
                       NULL,
                       sizeof(OBJECT_DIRECTORY),
                       0,
                       0,
                       (PVOID *)&PipeNameSpace);
       create_handle(NULL,
                       PipeNameSpace,
                       0,
                       ObjectAttributes.Attributes & OBJ_INHERIT,
                       (PHANDLE)&dev->pipes);

        if (!dev->pipes || !(dev->fd = alloc_pseudo_fd( &named_pipe_device_fd_ops, &dev->obj, 0 ))) {
            release_object( dev );
            dev = NULL;
        }
    }
    if (dev)
        make_object_static( &dev->obj );
}

extern int uk_poll(struct pollfd *ufds, int nfds, int timeout);

static int pipe_data_remaining( struct pipe_server *server )
{
    struct pollfd pfd;
    int fd;

#if 0
    assert( server->client );
#endif

    fd = get_unix_fd( server->client->fd );
    if (fd < 0)
        return 0;
    pfd.fd = fd;
    pfd.events = POLLIN;
    pfd.revents = 0;

    if (0 > uk_poll( &pfd, 1, 0 ))
        return 0;
 
    return pfd.revents&POLLIN;
}

static void check_flushed( void *arg )
{
    struct pipe_server *server = (struct pipe_server*) arg;

#if 0
    assert( server->event );
#endif
    if (pipe_data_remaining( server )) {
        server->flush_poll = add_timeout_user( -TICKS_PER_SEC / 10, check_flushed, server );
    }
    else {
        /* notify_empty( server ); */
        server->flush_poll = NULL;
        set_event( server->event, EVENT_INCREMENT, FALSE );
        release_object( server->event );
        server->event = NULL;
    }
}

static void pipe_server_flush( struct fd *fd, struct kevent **event )
{
    struct pipe_server *server = get_fd_user( fd );

    if (!server || server->state != ps_connected_server)
        return;

    /* FIXME: if multiple threads flush the same pipe,
              maybe should create a list of processes to notify */
    if (server->flush_poll)
        return;

    if (pipe_data_remaining( server )) {
        /* this kind of sux -
           there's no unix way to be alerted when a pipe becomes empty */
        server->event = create_event( NULL, NULL, 0, 0, 0, NULL );
        if (!server->event)
            return;
        server->flush_poll = add_timeout_user( -TICKS_PER_SEC / 10, check_flushed, server );
        *event = server->event;
    }
}

static void pipe_client_flush( struct fd *fd, struct kevent **event )
{
    /* FIXME: what do we have to do for this? */
}

#define FILE_SYNCHRONOUS_IO_ALERT       0x00000010
#define FILE_SYNCHRONOUS_IO_NONALERT    0x00000020

static inline int is_overlapped( unsigned int options )
{
    return !(options & (FILE_SYNCHRONOUS_IO_ALERT | FILE_SYNCHRONOUS_IO_NONALERT));
}

static enum server_fd_type pipe_server_get_fd_type( struct fd *fd )
{
    return FD_TYPE_PIPE;
}

static enum server_fd_type pipe_client_get_fd_type( struct fd *fd )
{
    return FD_TYPE_PIPE;
}

static obj_handle_t alloc_wait_event( struct w32process *process )
{
    obj_handle_t handle = 0;
    struct kevent *event = create_event( NULL, NULL, 0, 1, 0, NULL );

    if (event) {
        handle = alloc_handle( process, event, EVENT_ALL_ACCESS, 0 );
        release_object( event );
    }
    return handle;
}

static obj_handle_t pipe_server_ioctl( struct fd *fd, ioctl_code_t code, const async_data_t *async_data,
                                       const void *data, data_size_t size )
{
    struct pipe_server *server = get_fd_user( fd );
    struct async *async;
    obj_handle_t wait_handle = 0;

    switch(code) {
    case FSCTL_PIPE_LISTEN:
        switch(server->state) {
        case ps_idle_server:
        case ps_wait_connect:
            if (!async_data->event && !async_data->apc) {
                async_data_t new_data = *async_data;
                if (!(wait_handle = alloc_wait_event( get_current_w32process() )))
                    break;
                new_data.event = wait_handle;
                if (!(async = fd_queue_async( server->ioctl_fd, &new_data, ASYNC_TYPE_WAIT, 0 ))) {
                    close_handle( get_current_eprocess(), wait_handle );
                    break;
                }
            }
            else async = fd_queue_async( server->ioctl_fd, async_data, ASYNC_TYPE_WAIT, 0 );

            if (async) {
                set_server_state( server, ps_wait_open );
                if (server->pipe->waiters)
                    async_wake_up( server->pipe->waiters, STATUS_SUCCESS );
                release_object( async );
                set_error( STATUS_PENDING );
                return wait_handle;
            }
            break;
        case ps_connected_server:
            set_error( STATUS_PIPE_CONNECTED );
            break;
        case ps_disconnected_server:
            set_error( STATUS_PIPE_BUSY );
            break;
        case ps_wait_disconnect:
            set_error( STATUS_NO_DATA_DETECTED );
            break;
        case ps_wait_open:
            set_error( STATUS_INVALID_HANDLE );
            break;
        }
        return 0;

    case FSCTL_PIPE_DISCONNECT:
        switch(server->state) {
        case ps_connected_server:
#if 0
            assert( server->client );
            assert( server->client->fd );
#endif

            notify_empty( server );

            /* dump the client and server fds, but keep the pointers
               around - client loses all waiting data */
            do_disconnect( server );
            set_server_state( server, ps_disconnected_server );
            break;
        case ps_wait_disconnect:
#if 0
            assert( !server->client );
#endif
            do_disconnect( server );
            set_server_state( server, ps_wait_connect );
            break;
        case ps_idle_server:
        case ps_wait_open:
            set_error( STATUS_PIPE_LISTENING );
            break;
        case ps_disconnected_server:
        case ps_wait_connect:
            set_error( STATUS_PIPE_DISCONNECTED );
            break;
        }
        return 0;

    default:
        return default_fd_ioctl( fd, code, async_data, data, size );
    }
}

extern void clear_error(void);

static struct named_pipe *create_named_pipe( obj_handle_t root, const struct unicode_str *name,
                                             unsigned int attr )
{
    struct named_pipe *pipe = NULL;

    if (!name || !name->len)
        return alloc_wine_object( &named_pipe_ops );

    pipe = create_wine_object(root, &named_pipe_ops, name, NULL);
    if (pipe)
        clear_error();
    return pipe;
}

static struct pipe_server *get_pipe_server_obj( struct w32process *process,
                                obj_handle_t handle, unsigned int access )
{
    struct object *obj;
    obj = get_wine_handle_obj( process, handle, access, &pipe_server_ops );
    return (struct pipe_server *) obj;
}

static struct pipe_server *create_pipe_server( struct named_pipe *pipe, unsigned int options )
{
    struct pipe_server *server;

    server = alloc_wine_object( &pipe_server_ops );
    if (!server)
        return NULL;

    server->fd = NULL;
    server->pipe = pipe;
    server->client = NULL;
    server->flush_poll = NULL;
    server->options = options;

    list_add_head( &pipe->servers, &server->entry );
    grab_object( pipe );
    if (!(server->ioctl_fd = alloc_pseudo_fd( &pipe_server_fd_ops, &server->obj, options ))) {
        release_object( server );
        return NULL;
    }
    set_server_state( server, ps_idle_server );
    return server;
}

static struct pipe_client *create_pipe_client( unsigned int flags )
{
    struct pipe_client *client;

    client = alloc_wine_object( &pipe_client_ops );
    if (!client)
        return NULL;

    client->fd = NULL;
    client->server = NULL;
    client->flags = flags;

    return client;
}

static struct pipe_server *find_available_server( struct named_pipe *pipe )
{
    struct pipe_server *server;

    /* look for pipe servers that are listening */
    LIST_FOR_EACH_ENTRY( server, &pipe->servers, struct pipe_server, entry ) {
        if (server->state == ps_wait_open)
            return (struct pipe_server *)grab_object( server );
    }

    /* fall back to pipe servers that are idle */
    LIST_FOR_EACH_ENTRY( server, &pipe->servers, struct pipe_server, entry ) {
        if (server->state == ps_idle_server)
            return (struct pipe_server *)grab_object( server );
    }

    return NULL;
}

extern int uk_socketpair(int d, int type, int protocol, int sv[]);
#define AF_UNIX     1   /* Unix domain sockets      */
#define PF_UNIX                    AF_UNIX
#define SOCK_STREAM 1
#define SOL_SOCKET                 0xffff
#define SO_SNDBUF                  0x1001
#define SO_RCVBUF                  0x1002
extern int uk_setsockopt(int s, int  level,  int  optname,  const  void  *optval, int optlen);

struct object *named_pipe_open_file( struct object *obj, unsigned int access,
                                            unsigned int sharing, unsigned int options )
{
    struct named_pipe *pipe = (struct named_pipe *)obj;
    struct pipe_server *server;
    struct pipe_client *client;
    int fds[2];

    if (!(server = find_available_server( pipe ))) {
        set_error( STATUS_PIPE_NOT_AVAILABLE );
        return NULL;
    }

    if ((client = create_pipe_client( options ))) {
        if (uk_socketpair( PF_UNIX, SOCK_STREAM, 0, fds ) >= 0) {
#if 0
            assert( !server->fd );
#endif

            /* for performance reasons, only set nonblocking mode when using
             * overlapped I/O. Otherwise, we will be doing too much busy
             * looping */
            if (is_overlapped( options )) uk_fcntl( fds[1], F_SETFL, O_NONBLOCK );
            if (is_overlapped( server->options )) uk_fcntl( fds[0], F_SETFL, O_NONBLOCK );

            if (pipe->insize) {
                uk_setsockopt( fds[0], SOL_SOCKET, SO_RCVBUF, &pipe->insize, sizeof(pipe->insize) );
                uk_setsockopt( fds[1], SOL_SOCKET, SO_RCVBUF, &pipe->insize, sizeof(pipe->insize) );
            }
            if (pipe->outsize) {
                uk_setsockopt( fds[0], SOL_SOCKET, SO_SNDBUF, &pipe->outsize, sizeof(pipe->outsize) );
                uk_setsockopt( fds[1], SOL_SOCKET, SO_SNDBUF, &pipe->outsize, sizeof(pipe->outsize) );
            }

            client->fd = create_anonymous_fd( &pipe_client_fd_ops, fds[1], &client->obj, options );
            server->fd = create_anonymous_fd( &pipe_server_fd_ops, fds[0], &server->obj, server->options );
            if (client->fd && server->fd) {
                if (server->state == ps_wait_open)
                    fd_async_wake_up( server->ioctl_fd, ASYNC_TYPE_WAIT, STATUS_SUCCESS );
                set_server_state( server, ps_connected_server );
                server->client = client;
                client->server = server;
            }
            else {
                release_object( client );
                client = NULL;
            }
        }
        else {
#if 0
            file_set_error();
#endif
            release_object( client );
            client = NULL;
        }
    }
    release_object( server );
    return &client->obj;
}

typedef struct _FILE_PIPE_WAIT_FOR_BUFFER {
    LARGE_INTEGER   Timeout;
    ULONG           NameLength;
    BOOLEAN         TimeoutSpecified;
    WCHAR           Name[1];
} FILE_PIPE_WAIT_FOR_BUFFER, *PFILE_PIPE_WAIT_FOR_BUFFER;
typedef signed __int64   INT_PTR, *PINT_PTR;
#define FIELD_OFFSET(type, field) (/*(LONG)(INT_PTR)&*/(((type *)0)->field))  /* D.M. TBD */

static obj_handle_t named_pipe_device_ioctl( struct fd *fd, ioctl_code_t code, const async_data_t *async_data,
                                             const void *data, data_size_t size )
{
    struct named_pipe_device *device = get_fd_user( fd );

    switch(code) {
    case FSCTL_PIPE_WAIT:
        {
            const FILE_PIPE_WAIT_FOR_BUFFER *buffer = data;
            obj_handle_t wait_handle = 0;
            struct named_pipe *pipe;
            struct pipe_server *server;
            struct unicode_str name;

            if (size < sizeof(*buffer) ||
                size < FIELD_OFFSET(FILE_PIPE_WAIT_FOR_BUFFER, Name[buffer->NameLength/sizeof(WCHAR)])) {
                set_error( STATUS_INVALID_PARAMETER );
                return 0;
            }
            name.str = buffer->Name;
            name.len = (buffer->NameLength / sizeof(WCHAR)) * sizeof(WCHAR);
            if (!(pipe = (struct named_pipe *)find_object( device->pipes, &name, OBJ_CASE_INSENSITIVE ))) {
                set_error( STATUS_PIPE_NOT_AVAILABLE );
                return 0;
            }
            if (!(server = find_available_server( pipe ))) {
                struct async *async;

                if (!pipe->waiters && !(pipe->waiters = create_async_queue( NULL ))) goto done;

                if (!async_data->event && !async_data->apc) {
                    async_data_t new_data = *async_data;
                    if (!(wait_handle = alloc_wait_event( get_current_w32process() ))) goto done;
                    new_data.event = wait_handle;
                    if (!(async = create_async( current_thread, pipe->waiters, &new_data ))) {
                        close_handle( get_current_eprocess(), wait_handle );
                        wait_handle = 0;
                    }
                }
                else async = create_async( current_thread, pipe->waiters, async_data );

                if (async) {
                    timeout_t when = buffer->TimeoutSpecified ? buffer->Timeout.QuadPart : pipe->timeout;
                    async_set_timeout( async, when, STATUS_IO_TIMEOUT );
                    release_object( async );
                    set_error( STATUS_PENDING );
                }
            }
            else release_object( server );

        done:
            release_object( pipe );
            return wait_handle;
        }

    default:
        return default_fd_ioctl( fd, code, async_data, data, size );
    }
}

static NTSTATUS
parse_named_pipe(
		IN PVOID ParseObject,
		IN PVOID ObjectType,
		IN OUT PACCESS_STATE AccessState,
		IN KPROCESSOR_MODE AccessMode,
		IN ULONG Attributes,
		IN OUT PUNICODE_STRING CompleteName,
		IN OUT PUNICODE_STRING RemainingName,
		IN OUT PVOID Context OPTIONAL,
		IN PSECURITY_QUALITY_OF_SERVICE SecurityQos OPTIONAL,
		OUT PVOID *Object
		)
{
    struct named_pipe_device *device = (struct named_pipe_device *)ParseObject;
    NTSTATUS status;

    status = ref_object_by_handle(
                    device->pipes,
                    NAMED_PIPE_ALL_ACCESS,
                    NULL,
                    KernelMode,
                    Object,
                    NULL);

    return status;
}

DECL_HANDLER(create_named_pipe)
{
    struct named_pipe *pipe;
    struct pipe_server *server;
    struct unicode_str name;
    struct directory *root = NULL;

    reply->handle = 0;
    get_req_unicode_str( &name );

    pipe = create_named_pipe( req->rootdir, &name, req->attributes | OBJ_OPENIF );

    if (root)
        release_object( root );
    if (!pipe)
        return;

    if (get_error() != STATUS_OBJECT_NAME_EXISTS) {
        /* initialize it if it didn't already exist */
        pipe->instances = 0;
        pipe->waiters = NULL;
        INIT_LIST_HEAD( &pipe->servers );
        pipe->insize = req->insize;
        pipe->outsize = req->outsize;
        pipe->maxinstances = req->maxinstances;
        pipe->timeout = req->timeout;
        pipe->flags = req->flags;
    }
    else {
        if (pipe->maxinstances <= pipe->instances) {
            set_error( STATUS_INSTANCE_NOT_AVAILABLE );
            release_object( pipe );
            return;
        }
        if ((pipe->maxinstances != req->maxinstances) ||
            (pipe->timeout != req->timeout) ||
            (pipe->flags != req->flags)) {
            set_error( STATUS_ACCESS_DENIED );
            release_object( pipe );
            return;
        }
        clear_error(); /* clear the name collision */
    }

    server = create_pipe_server( pipe, req->options );
    if (server) {
        reply->handle = alloc_handle( get_current_w32process(), server, req->access, req->attributes );
        server->pipe->instances++;
        release_object( server );
    }

    release_object( pipe );
}

DECL_HANDLER(get_named_pipe_info)
{
    struct pipe_server *server;
    struct pipe_client *client = NULL;

    server = get_pipe_server_obj( get_current_w32process(), req->handle, FILE_READ_ATTRIBUTES );
    if (!server) {
        clear_error();
        client = (struct pipe_client *)get_wine_handle_obj( get_current_w32process(), req->handle,
                                                       FILE_READ_ATTRIBUTES, &pipe_client_ops );
        if (!client)
            return;
        server = client->server;
    }

    reply->flags        = server->pipe->flags;
    reply->maxinstances = server->pipe->maxinstances;
    reply->instances    = server->pipe->instances;
    reply->insize       = server->pipe->insize;
    reply->outsize      = server->pipe->outsize;

    if (client)
        release_object(client);
    else {
        reply->flags |= NAMED_PIPE_SERVER_END;
        release_object(server);
    }
}
#endif /* CONFIG_UNIFIED_KERNEL */
